'''Map the source form parameters into a form that the etc_engine can use.

Two approaches are used. The old is to construct a synphot source
expression. The new is to construct a pysynphot spectrum object.
Both take a form dictionary as input.
'''
import os, sys

import pandeia.engine.helpers.bit.pyetc_util as util

#Set up some constants
wave_scaling = {'microns':    10.0**4,
                'nanometers': 10.0,
                'angstroms':   1.0}
symbol = {'microns': "&mu;",
          'nanometers': "nm",
          'angstroms': "&Aring"}

# Map of target choices to synphot files/expressions
SDICT = util.read_dict(os.path.join(os.path.dirname(__file__),'etc_web_data/targets.dat'))

# Map target kind onto target value selector variable
SPECNAMEMAP = {
    'SpectrumCK':'fSpectrumCK',
    'SpectrumPickles':'fSpectrumPickles',
    'SpectrumKurucz':'fbpgsfile',
    'SpectrumStellar':'fStellar',
    'SpectrumNonStellar':'fnonstellar',
    'SpectrumHST': 'fcalfile',
    'SpectrumPhoenix': 'fphoenixfile'
    }

# TARGET_HELPERS massage the raw data of the target mapping file into description and
# pysynphot expression tuples,  basically user results text + psyn code.
TARGET_HELPERS = {
    # form_var    :  lambda key, target.dat[form_var][key] :   description,  psyn_expr
    "fSpectrumCK"      : lambda k, v:  ("Castelli-Kurucz Models " + k + " " + v,      'icat(ck04models,%s,0.0,%s)' % tuple(v.split())),
    "fbpgsfile"        : lambda k, v:  ("Kurucz Models " + k + " " + v,               'icat(k93models,%s,0.0,%s)' % tuple(v.split())),
    "fSpectrumPickles" : lambda k, v:  ("Pickles Models " + v[0],                         v[1]),
    "fcalfile"         : lambda k, v:  ("Standard HST Star Spectra " + v[0],          "spec(%s)" % (v[1],)),
    "fphoenixfile"     : lambda k, v:  ("Phoenix M dwarf Models " + v[0],             "spec(%s)" % (v[1],)),
    "fStellar"         : lambda k, v:  ("Bruzual Synthetic Stellar Spectra " + v[0],  "spec(%s)" % v[1]),
    "fnonstellar"      : lambda k, v:  ("Non-stellar Objects " + v[0],                "spec(%s)" % v[1]),
}

# Provides filter names for use in building output page. Neither .capitalize()
# or .upper() work as a generic solution; we need to apply one of them and
# counteract exception cases with this dictionary.
filter_output_names = {"SDSS":   "Sloan",
                       "ACS":    "ACS",
                       "NICMOS": "NICMOS",
                       "WFC3":   "WFC3"}

def map_filter(filter):
    '''convert the web form values for the normalization bandpass into a
    form that synphot/pysynphot can use'''
    #Normally, we can just replace a / by a comma and we're done.
    filter = filter.lower().replace('/',',')

    #But we have to change sloan to sdss
    if filter.startswith('sloan'):
        filter = filter.replace('sloan','sdss')

    #and specify the detector for NICMOS.
    #JETC uses nicmos, 2: ETC/datafiles/SynphotBands.xml, lines 110:123
    elif filter.startswith('nicmos'):
        filter = filter.replace('nicmos','nicmos,2')

    # and for WFC3UVIS
    elif filter.startswith('wfc3'):
        filter = filter.replace('uvis','uvis1')

    # the ACS detector is not specified; we arbitrarily pick WFC
    elif filter.startswith('acs'):
        filter = filter.replace('acs','acs,wfc1')
    return filter

def map_name(stype, fname):
    '''determine specific file matching given name or key

    >>> map_name('SpectrumPickles', 'O5V')
    ('Pickles O5V', 'crgrid$pickles/dat_uvk/pickles_uk_1.fits')
    '''
    try:                              # .e.g. stype=SpectrumPickles
        keyvar = SPECNAMEMAP[stype]   # e.g.  keyvar=fSpectrumPickles
        return TARGET_HELPERS[keyvar](fname, SDICT[keyvar][fname])
    except KeyError:
        raise ValueError("Unrecognized source type: %s, %s"%(stype,fname))

class Source(object):
    """
    This class embodies the source descriptions and abiltiy to construct
    pysynphot spectrum objects.
    """
    def __init__(self, request_dir, formdata=None, obsmode=None):
        """Construct source object"""
        self.request_dir = request_dir
        self.obsmode=obsmode
        if formdata:
            self.inputs_from_form(formdata)
        else:
            self.expr = ''
            self.help = '' #for output web form

    def inputs_from_form(self, formdata):
        '''Construct a synphot source expression from the form info.
        Currently forms consist of mostly exclusive options in dict
        format'''

        # These functions construct both the synphto expression element
        # and the user-friendly string displayed on web output.

        # specexpr is the fundamental source expression, which can
        # be redshifted, extincted, renormalized, and had emission lines
        # added to it. The order of these operations matters.
        spechelp, specexpr = self.handle_source(formdata)
        ehelp, estr        = self.handle_extinction(formdata)
        rnhelp, rnstr      = self.handle_renormalization(formdata,obsmode=self.obsmode)
        zhelp, zstr     = self.handle_redshift(formdata)
        lhelp, lines    = self.handle_emission_lines(formdata)

        #First apply redshift
        specexpr = zstr % specexpr

        #Then apply extinction either before or after normalization
        if formdata['fextinctiontype'] == 'before':
            specexpr += estr
            specexpr = rnstr % specexpr
        else:
            specexpr = rnstr % specexpr
            specexpr += estr

        #Finally, assign to the expr attribute.
        #Lines will be passed in separately.
        self.expr = specexpr



        #Also construct the "help" attribute for user friendliness.
        #Note emission lines returns non-standard multi-line help

        #TODO: Make pysyn_help display smarter so it only shows when
        #mode=debug; or, filter out filenames. In the meantime, disable
        #this display for security reasons.
        pysyn_help = ["Pysynphot:", self.expr]
        self.help = [spechelp, ehelp, rnhelp, zhelp] + lhelp #+ [["<p/>",""], pysyn_help]

        #Verify that we got a result out of all that work before returning.
        if not self.expr:
            #Lines might be specified separately
            if lines:
                #Keep pysynphot parser from failing: specify flat spectrum
                #that is uniformly zero
                self.expr = 'unit(0.0,flam)'
            else:
                raise ValueError('Source form selections did not result in a valid source specification.\n These might include no flux from the selected source combination.')

    def handle_source(self, fd):
        # Always uses key: fsorigin
        # Conditionally uses one of:
        #  fSpectrumCK, fSpectrumPickles, fbpgsfile, fStellar, fcalfile, fphoenixfile, fnonstellar,fbbtemp,(fIndex,fIsLambda)
        # determine which option to use
        stype = fd['fsorigin']
        if stype == 'SpectrumUpload':
            specexpr = fd['fUploadFileLocal']
            if self.request_dir not in [".","./"]:
                specexpr = self.request_dir + '/' + specexpr
            info = "User uploaded spectrum:  " + fd['fUploadFile']
        elif stype == 'SpectrumHstOther':
            specexpr = fd['fOtherFile']
            info = "Other HST spectrum " + repr(fd['fOtherFile'])
        elif stype in ['SpectrumCK', 'SpectrumPickles','SpectrumKurucz','SpectrumStellar',
                       'SpectrumNonStellar','SpectrumHST', 'SpectrumPhoenix']:
            selected = fd[SPECNAMEMAP[stype]]
            info, specexpr = map_name(stype, selected)

        #The following options have sanity checking applied to them:
        elif stype == 'SpectrumBlackBody':
            kelvins = float(fd['fbbtemp'])
            if kelvins <= 0:
                raise ValueError("Temperature %f must be positive"%kelvins)
            specexpr = 'bb(%f)' % kelvins
            info = "Black body at " + str(kelvins) + "K"

        elif stype == 'SpectrumPowerLaw':
            specexpr = 'pl(1.,%f,flam)' % fd['fIndex']
            info = "Power Law with index " + str(fd['fIndex'])
        elif stype == 'SpectrumFlat':
            units = ['flam','fnu'][int(fd['fIsLambda']=='false')]
            specexpr = 'unit(1.,%s)' % units
            info = "Flat in units of " + {"flam" : "F&lambda;",
                                          "fnu"  : "F&nu;", } [units]
        elif stype == 'SpectrumEmpty':
            specexpr = ''
            info = "Empty"
        else:
            raise ValueError({},'Unknown option %s for source spectrum'%stype)
        return ["Spectrum:",info], specexpr

    EXTINCTION_MAP  = {
        "mwavg"   : "Milky Way Diffuse (Rv=3.1)",
        "mwdense" : "Milky Way Dense (Rv=5.0)",
        "mwrv21"  : "Milky Way CCM (Rv=2.1)",
        "mwrv4"   : "Milky Way CCM (Rv=4.0)",
        "lmcavg"  : "LMC Diffuse (Rv=3.41)",
        "lmc30dor": "LMC 30DorShell (Rv=2.76)",
        "smcbar"  : "SMC Bar (Rv=2.74)",
        "xgalsb"  : "Starburst (attenuation law)",
        }

    def handle_extinction(self, fd):
        # extinction model section
        #Always uses key: febv
        #Conditionally uses key: febmvtype
        ebv = fd['febv']
        if ebv:
            etype = fd['febmvtype']
            estr = ' * ebmvx(%f,%s)' % (ebv, etype)
            ehelp = self.EXTINCTION_MAP[ etype ] + " = " + str(ebv)
            ehelp += " applied " + fd['fextinctiontype'] + " normalization"
        else:
            estr = ''
            ehelp = 'None'
        return ["Extinction E(B-V):", ehelp], estr

    def handle_redshift(self, fd):
        # redshift handling
        # Always uses key: fRedshift
        z = fd['fRedshift']
        if float(z) < 0:
            raise ValueError("Redshift %f must be positive"%z)
        if z:
            zstr = 'z(%s' + ',%f)' % z
            zhelp = "z=" + str(z)
        else:
            zstr = '%s'
            zhelp = "None"
        return ["Redshift:", zhelp], zstr

    def _build_bandpass(self, fd):

        instrument = fd['instrument']
        science_mode = fd['science_mode']

        module_name = 'pandeia.engine.helpers.bit.instruments.hst.%s.web.%s_web' % (instrument, instrument)
        make_obsmode_function_name = 'make_%s_%s_obsmode' % (instrument, science_mode)

        module = util.dynamic_import(module_name)
        make_obsmode_function = getattr(module, make_obsmode_function_name)

        # STIS needs special handling: the interface to the function
        # that builds the obsmode string is non-standard and depends
        # on the science mode.
        if instrument == 'stis':
            # we delegate to the instrument_from_form() method the
            # details on how to process the input in a way that
            # makes sense to the obsmode building function.
            instrument_from_form_function_name = 'instrument_from_form'
            instrument_from_form_function = getattr(module, instrument_from_form_function_name)
            idict  = instrument_from_form_function(science_mode, fd)

            if science_mode == 'imaging':
                results = make_obsmode_function(idict['detector'], idict['aperture'])
            elif science_mode == 'targetacquisition':
                results = make_obsmode_function(idict['detector'],
                                                idict['aperture'],
                                                idict['ccdmode'],
                                                (fd['fsourceType'] == 'point'))
            else: # spectroscopic
                results = make_obsmode_function(fd)

        else:
            results = make_obsmode_function(fd)

        # the XXX_web.py API has a bit of internal variation....
        if type(results) == str:
            return results
        else:
            return results[0]

    def handle_renormalization(self, fd, obsmode=None):
        # renormalization section
        # Always uses key: fnormalize
        # Conditionally uses key: fftype_filters, rn_flux_bandpass, rn_flux_bandpass_units,
        rntype = fd['fftype']
        # What we're doing here is building a string into which the
        # spec_expr will be substituted. This is to support the option
        # of applying the extinction either before or after the
        # renormalization.
        stype = fd['fsorigin']

        if (rntype == 'fno' or stype == 'SpectrumEmpty'): #cannot renormalize nothing
            rnstr = '%s'
            rnhelp = "None"

        elif rntype.startswith('fnormalize'):

            # pick off the normalization mode from the rntype string
            junk, norm_mode = rntype.split('.',1)

            # normalize at a specific wavelength.
            if norm_mode == 'lambda':
                fflux = fd['rn_flux_lambda']
                filter = fd['rn_flux_lambda_units']

                flambda = fd['rn_lambda']
                lambda_units = fd['rn_lambda_units']
                flambda_angstrom = flambda * wave_scaling[lambda_units]

                # Bandpass is box 1.0 A wide.
                rnstr = 'rn(%s,' +'box(%f,1.),%g,%s)' % (flambda_angstrom, fflux, filter)
                rnhelp = "Flux " + str(fflux) + " " + str(filter) + " at " + \
                         str(flambda) + " " + symbol[lambda_units]

            # normalize by filter or bandpass
            else:
                fflux = fd['rn_flux_bandpass']
                funits = fd['rn_flux_bandpass_units']
                filter_type = fd['fftype_filters'].split('.',1)[1]

                if filter_type == 'observation':
                    # normalize by bandpass used for observation
                    fband = self._build_bandpass(fd)
                    band_type = " bandpass defined above."

                else:
                    # normalize by user-chosen filter
                    filter_selector = "filter." + filter_type
                    fband = map_filter(fd[filter_selector])

                    # good-looking output is tricky.
                    fband_output = fband.upper().replace(',', '/')
                    tokens = fband_output.split('/')
                    tokens[0] = filter_output_names.get(tokens[0], tokens[0].capitalize())
                    band_type = " filter " + '/'.join(tokens)

                rnstr = 'rn(%s,' + 'band(%s),%g,%s)' % (fband, fflux, funits)
                rnhelp = "Renormalized to " + funits + " = " + str(fflux) + " in " + band_type

        else:
            raise ValueError('Unrecognized normalize option %s'%rntype)
        return ["Normalization:", rnhelp], rnstr

    def handle_emission_lines(self, fd):
        """ Produce a list of dicts that specify the lines.
        See engine/observable.ObservableItem docstring for
        details.
        """
        # emission line section

        #Central wavelength may be either in microns or angstroms.
        #FWHM is always in angstroms
        #We will convert both to angstroms for pysynphot.
        self.lines = list()
        waveunits = fd.get('wavelengthUnits','angstroms')
        wavesymbol = symbol[waveunits]
        scl = wave_scaling[waveunits]
        info = [["Emission Lines:",""]]
        #Up to three lines
        for n in (1,2,3):
            c, w, f = fd['fL%d_center'%n], fd['fL%d_fwhm'%n], fd['fL%d_flux'%n]
            if c*w*f > 0:
                self.lines.append(dict(center = c*scl,
                                       fwhm = w,
                                       flux = f,
                                       waveunits = 'angstroms',
                                       fluxunits = 'flam')
                                  )
                info.append(["Line at", "Center " + str(c) + " " + wavesymbol + "&nbsp;&nbsp;&nbsp; "
                             "FWHM " + str(w) + " &Aring;&nbsp;&nbsp;&nbsp;"
                             "Flux " + str(f) + " F&lambda;"])

        if not self.lines:
            self.lines = None
            info = [["Emission Lines:", "None"]]

        return [info, self.lines]



from pandeia.engine.helpers.background.jwst import sl_accessor_funcs as funcs


def get_stray_light_bg(ra, dec, date, ra_dec_str, date_str):

    return funcs.get_sl(ra, dec, date, ra_dec_str, date_str)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Module that implements a background noise calculator (BNC) for the pandeia ETC engine
"""

import numpy as np
from warnings import warn

from .scene import Scene
from .custom_exceptions import EngineInputError
from .calc_utils import build_empty_scene
from .etc3D import setup, make_observation

from . import debug_utils


def bnc(calc_input, validate=False):
    """
    This function uses the pandeia ETC engine to calculate the standard deviation of the per-pixel count rate for a
    given calculation when only the background signal is included. The input can be any valid pandeia calculation.
    The scene information will be ignored and the strategy information ignored except for number of dithers. The output
    will be the maximum standard deviation (sqrt(variance)) encountered at the detector plane.

    Parameters
    ----------
    calc_input: dict
        Pandeia engine input API compliant dict containing information required to perform the background noise calculation

    Returns
    -------
    pix_stddev: float
        Maximum standard deviation in pixel count rate (e-/sec/pixel) generated by the input background signal.
    """

    calc_config, instrument, strategy, scene_configuration, background, background_level, warnings = setup(calc_input, validate=validate)

    # ignore the scene in the input calculation and use an empty one instead. the only input signal in the calculation will
    # be the background which is spatially constant across the scene.
    scene_configuration = build_empty_scene()

    debug_utils.init(calc_input)

    scene = Scene(instrument.telescope.tel_name, config=scene_configuration)
    
    # dithering images uncorrelates all of the noise and thus reduces the variance by ndithers:
    #   variance = variance_single/ndithers
    if hasattr(strategy, 'dithers'):
        ndithers = len(strategy.dithers)
        strategy.dithers = [{'x': 0.0, 'y': 0.0}] # we don't actually need or want to calculate all of them now
    else:
        ndithers = 1
    my_detector_signal_list, my_detector_noise_list, my_detector_saturation_list, obs = make_observation(calc_config, instrument, 
                                                                       strategy, scene, background, background_level, validate=validate)

    # get the detector plane noise products
    # dithers have the same setup, just a different pointing, so we can be confident using only the first one
    if validate:
        det_var, det_stddev, det_rn_var, det_ff_var = my_detector_noise_list[0].on_detector()
    else:
        det_var, det_stddev, det_rn_var = my_detector_noise_list[0].on_detector()

    pix_stddev = det_stddev.max() / np.sqrt(ndithers)
    return pix_stddev



# STIS  limits

# This file is read when the instruments/data.py file is executed
# as part of the instrument module import. The resulting dictionary
# is used to initialize instances of Limit that will be used to
# check any subsequent STIS calculation.

# allsub defines what the magic value "all" should expand into
allsub = {'mode':['imaging', 'spectroscopic', 'targetacquisition'],
          'detector' : ['ccd', 'fuvmama', 'nuvmama']
         }

# limit_spec contains the actual limit definitions. It is a
# dictionary {limit shortname : {definition of limit} }.
# See user's guide for detailed requirements of the limit definition.
#
# Message strings may use the following variable names for substitution:
#  varname = value of the variable specified in "varname"
#  threshold = value specified in "threshold"
#  percentval = percentage expression of (1-subthresh)
#

#Pedigree:
#
# Most of these values were taken from the STIS IHB for Cycle 18 (version 9.0),
# Proffitt, C., et al.
#   STIS CCD Limits: Table 7.1 (page 112), under "Saturation limit"
#   STIS MAMA Limits: Table 7.8 (page 159) (the entire table
#           consists of MAMA bright-object limits).
#
# The CCD readnoise value for gain=4 was updated to match the value in the
# Cycle 19 Instrument Handbook (page 111) and to be consistent with current
# analysis.
#
# This value comes from plots made as part of the STIS Bias and Readnoise
# Monitor program. Plots can be found at:
# http://www.stsci.edu/hst/stis/calibration/Monitors/Read-Noise

limit_spec = {
    'brightpix' : {
    'varname' : 'brightest_pixel_counts',
    'mode': 'all',
    'varieties' : ( #ccd
       { 'detector' : 'ccd',
         'mode': 'all',
         'threshold' : ('gain', {
             1 : 33000,
             4 : 120000} ), #updated per pr#68967
         'message' : 'CCD Full Well limits: total counts per native pixel %(varname)g exceeds bright limit %(threshold)g for selected gain',
         'subthresh' : 0.7,
         'submessage' : 'Total counts per binned pixel %(varname)g is within %(percentval)d percent of the CCD fullwell limit %(threshold)g',
         },
       # fuvmama  imaging thresholds
       { 'detector' : 'fuvmama',
         'varname'  : 'brightest_pixel_rate',
         'mode': 'imaging',
         'threshold' : 100,
         'message' : 'MAMA Bright limits: count rate of brightest pixel %(varname)g exceeds bright limit %(threshold)g',},
       # fuvmama spec thresholds
       { 'detector' : 'fuvmama',
         'varname'  : 'brightest_pixel_rate',
         'mode'     : 'spectroscopic',
         'threshold': 75,
         'message' :  'MAMA Bright limits: count rate of brightest pixel %(varname)g exceeds bright limit %(threshold)g,',},
       # nuvmama imaging thresh
       { 'detector' : 'nuvmama',
         'varname'  : 'brightest_pixel_rate',
         'mode'     : 'imaging',
         'threshold': 100,
         'message'  : 'MAMA Bright limits: count rate of brightest pixel %(varname)g exceeds bright limit %(threshold)g',},
       # nuvmama spec thresh
       { 'detector' : 'nuvmama',
         'varname' : 'brightest_pixel_rate',
         'mode'    : 'spectroscopic',
         'threshold' : 75,
         'message' : 'MAMA Bright limits: count rate of brightest pixel %(varname)g exceeds bright limit %(threshold)g',},
       ) #end varieties
   } , #end brightpix

    'brightimg': {
        'varname': 'detector_total_rate',
        'detector': ['nuvmama','fuvmama'],
        'message' : 'MAMA Bright limits: total count rate per image %(varname)g exceeds bright limit %(threshold)g',
        'subthresh' : 0.4,
        'submessage': 'MAMA Bright limits: total count rate per image %(varname)g exceeds limit for irregularly-variable sources %(percentval)3g percent of the bright limit %(threshold)g counts per second',
        'varieties': (

        # fuvmama/nuvmama imaging
        {'mode'     : 'imaging',
         'threshold': 200000,},

        #fuvmama/nuvmama spec, first order modes
        {'mode'      : 'spectroscopic',
         'threshold' : ('grating', 
                        #This is the set of "first order modes"
                        {# NUVMAMA
                         'g230l' : 30000,  
                         'g230m' : 30000,
                         # FUVMAMA
                         'g140l' : 30000,
                         'g140m' : 30000,

                        # This is the set of "other modes"
                          # NUVMAMA
                         'e230h': 200000,
                         'e230m': 200000,
                         'prism': 200000,
                         # FUVMAMA
                         'e140m': 200000,
                         'e140h': 200000,
                         }),
         }, #end spec
        ), #end varieties
     }, #end brightimg

    'brightbinpix': {
        'varname': 'brightest_pixel_counts',
        'detector': 'ccd',
        'mode': 'all',
        'message' : 'CCD Full Well limits: total counts per binned pixel %(varname)g exceeds bright limit %(threshold)g for selected gain',
        # these threshold values are placeholders only, so the code won't crash.
        # Actually the true threshold values are computed by a hard-coded formula in the code, in stis.py
        'threshold' : ('gain', {
            1 : 34300,
            4 : 65536} ),
     }, #end brightbinpix

    'minbackground' : {
        'varname' : 'background_pixel_counts',
        'mincheck' : True,
        'message' : '''"Electrons per pixel due to background (%(varname).2g) is less than the recommended
                        threshold of %(threshold)g electrons to avoid poor charge transfer efficiency (CTE).
                        We suggest you consider CTE mitigation strategies described in the STIS Instrument
                        Handbook."''',
        'mode' : ['imaging', 'spectroscopic'],
        'detector' : 'ccd',
        'per_read' : True,
        'threshold' : 20.0
        }, #end minbackground

    'mintime' : {
        'varname' : 'time',
        'message' : '''Result of the calculation %(varname)g is less than the minimum
                    exposure time for this detector (%(threshold)g) ''',
        'mincheck' : True,
        'mode' : 'all',
        'detector' : 'ccd',
        'per_read' : True,
        'threshold': 0.1,
       }, # end mintime

    'maxtime' : {
        'varname' : 'time',
        'message' : '''Result of the calculation %(varname)g is more than the maximum
                    exposure time for this detector (%(threshold)g) ''',
        'mode' : 'all',
        'detector' : 'ccd',
        'per_read' : True,
        'threshold': 16920.,
       }, #end maxtime

    'maxleak' : {
        'varname' : 'total_leak',
        'message' : '''This observation has a significant filter leak where
                       %(varname).3g percent of the signal comes from outside the prescribed
                       filter bandpass limits.''',
        'mode' : 'imaging',
        'detector' : 'all',
        'per_read' : False,
        'threshold' : 5.0,
        }, # end maxleak

    'zerotargetflux' : {
        'varname' : 'target_total_rate',
        'mincheck' : True,
        'mode' : 'imaging',
        'detector' : 'all',
        'per_read' : False,
        'threshold' : 1.E-25,
        'varieties': (
            {'mode'   : 'imaging',
             'message': '''Zero target flux in the observation bandpass''',},
            {'mode'   : 'spectroscopic',
             'message': '''Zero target flux at the observation wavelength''',},
        ),
       }, # end zerotargetflux

    'minbuftime' : {
       'varname' : 'buffer_time',
       'mincheck': True,
       'mode'    : ['spectroscopic','imaging'],
       'detector': ['nuvmama','fuvmama'],
       'threshold' : 99,
       'varieties' : (
         {
          'mode' : 'spectroscopic',
          'message': 'Buffer time %(varname)g is less than minimum %(threshold)g seconds.',},
         {'mode' : 'imaging',
          'message' : '''Because buffer time is less than %(threshold)g sec, this observation can
                         only be performed in ACCUM mode.''',
          }
          )
    }, #end minbuftime

} #end limit_spec




# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import copy
import astropy
import pkg_resources
from . import io_utils as io
from .utils import merge_data, apply_default_only
from .custom_exceptions import DataConfigurationError
from .pandeia_warnings import standard_warning_messages as warning_messages


default_refdata_directory = os.environ.get("pandeia_refdata")


def default_refdata(directory=None):
    """
    There is a default refdata that is either refdata/ (for convenience
    of a developer) or the value of the environment variable $pandeia_refdata
    After importing this module, you can call this function to explicitly
    set the name of the refdata directory
    """
    global default_refdata_directory
    if directory is not None:
        default_refdata_directory = directory
    return default_refdata_directory


class DefaultConfig:

    """
    This class provides functionality to discover and load defaults from JSON
    configuration files, update the configuration information from a dict and any kwargs
    passed from the caller, and then populate the objects attributes from the resulting dict.

    Parameters
    ----------
    config: dict (optional)
        dictionary containing necessary configuration information
    **kwargs: list of keyword/value pairs
        parameter keyword and value pairs to augment defaults and config
    """

    def __init__(self, config={}, webapp=False, **kwargs):
        # grab info from the configured defaults file, if any, from caller via a passed dict, or via keywords
        # clean meta blocks from the configuration files
        all_config = merge_data(self._get_config(), config, dict(**kwargs))
        # checking if all_config matches self._get_config() is preferable, because it handles both the case where 
        # config={} was either empty and where it had just a single entry equivalent to the default.
        if all_config == self._get_config() and webapp==False:
            all_config=apply_default_only(all_config)
        all_config = self._remove_meta(all_config)
        all_config = self._remove_default_only(all_config)
        self.warnings = {}
        # add configuration items to the class as attributes
        self.__dict__.update(all_config)

        # do some API checks
        if webapp:
            try:
                all_config = merge_data(config, dict(**kwargs))
                self._api_checks(all_config)
            except AttributeError as e:
                self.warnings['no_api_check'] = warning_messages['no_api_check'] % (self.__class__.__name__, e)

    def _remove_meta(self, config):
        """
        The meta block in our json files is solely used for comments, data pedigree, and
        other metadata. As such, nothing in the engine should read it or need it; removing
        meta should do nothing to calculations but reduce the possibility of errors when
        looping through a dictionary (#3791).

        This routine (recursively) removes meta blocks from the configuration

        Parameters
        ----------
        config: dict
            A fully populated dictionary, ready to be added to the attributes of whatever
            class called this function

        Returns
        -------
        config: dict
            A fully populated dictionary, cleaned of meta blocks
        """
        if 'meta' in config:
            del config['meta']
        for item in config:
            # if the item itself is a dictionary, recurse inside for a meta block
            if isinstance(config[item], dict):
                config[item] = self._remove_meta(config[item])

        return config

    def _remove_default_only(self, config):
        """
        The default_only value construct is used to denote values that should be added only when building a 
        calculation from scratch. In all other cases, we can assume the user, webapp, or test generator has 
        already specified the optional (or one of many, if there are many keywords that control similar 
        functionality) keyword(s), and therefore do not need to load it here as it might conflict with 
        the user's choice.

        This routine (recursively) removes default_only blocks from the configuration

        Parameters
        ----------
        config: dict
            A fully populated dictionary, ready to be added to the attributes of whatever class called 
            this function

        Returns
        -------
        config: dict
            A fully populated dictionary, cleaned of default_only blocks
        """
        if 'default_only' in config:
            del config['default_only']
        for item in config:
            # if the item itself is a dictionary, recurse inside for a default block
            if isinstance(config[item], dict):
                config[item] = self._remove_default_only(config[item])

        return config

    def _load_dict(self, config, keyword, msg="No default configuration found."):
        """
        The data-engine API is set up to load defaults from two sources: a file in
        $pandeia_refdata with configured information ABOUT the instruments, spectra, and
        capabilities; and the engine inputs that CHOOSE a particular setup supplied
        (directly, through engine scripting, or indirectly via the webapp) from the user.
        Both kinds of data are necessary (simultaneously, even) for the engine to
        function.

        It is, however, a common pattern to need to load more specific configuration data
        from $pandeia_refdata based on user inputs; for instance, the appropriate
        attributes for a given type of normalization bandpass or the parameters for a
        particular kind of analytic spectrum. In addition, if webapp=False (or when
        constructing a calculation input from scratch), an entire default set of choices
        should be selected.

        This is a common loading pattern.

        Parameters
        ----------
        config: dict
            Configuration dictionary from data
        """
        if keyword in config:
            keydata = config.pop(keyword)
            config.update(keydata)
        else:
            raise DataConfigurationError(value=msg)

        return config

    def as_dict(self):
        """
        Return dict representation of instance configuration. If self.api_parameters exists,
        use it to determine what to return.  otherwise, just use self.__dict__ to provide everything.

        Also scrub astropy quantities.
        """
        if hasattr(self, "api_parameters"):
            d = {}
            for p in self.api_parameters:
                d[p] = getattr(self, p)
        else:
            d = self.__dict__
        for item in d:
            if type(d[item]) is astropy.units.quantity.Quantity or type(d[item]) is astropy.units.function.logarithmic.Magnitude:
                d[item] = d[item].value
            if type(d[item]) in [astropy.units.core.PrefixUnit, astropy.units.core.Unit]:
                d[item] = d[item].to_string()
        return d

    def _get_config(self):
        """
        Read default configuration from JSON

        Returns
        -------
        config: dict
            All desired class attributes should have defaults defined in the config file
        """
        # use this trick to key the configuration file name off the name of the instantiated subclass
        objname = self.__class__.__name__.lower()
        config_file = "defaults/%s_defaults.json" % objname
        config_path = pkg_resources.resource_filename(__name__, config_file)
        config = io.read_json(config_path, raise_except=True)
        return config

    def _api_checks(self, conf):
        """
        Check input configuration against self.api_parameters to make sure all the expected
        parameters are being set and no unrecognized parameters are being passed in.

        Parameters
        ----------
        conf: dict
            Engine input API format dict
        """
        # first, go through self.api_parameters and make sure they're all there
        for p in self.api_parameters:
            if p not in conf:
                msg = warning_messages["missing_api"].format(
                    p,
                    self.__class__.__name__,
                    getattr(self, p)
                )
                self.warnings["{:s}_{:s}".format(self.__class__.__name__.lower(), p)] = msg

        all_pars = copy.deepcopy(self.api_parameters)
        if hasattr(self, "api_ignore"):
            all_pars.extend(self.api_ignore)

        # now go through the input and flag anything we don't know about
        for c in conf:
            if c not in all_pars and "ui_" not in c:
                msg = warning_messages["unsupported_api"].format(c, self.__class__.__name__)
                self.warnings["{:s}_unsupported_{:s}".format(self.__class__.__name__.lower(), c)] = msg



""" This module contains functions specific to scan mode.

"""

from ...engine import initialize
from ...engine.area_over_which import ImageRectangularArea
from ...engine.string_constants import SNR_REGION, SPECTROSCOPIC, SCSPECTROSCOPIC, IMAGING, SCIMAGING


scan_science_modes = {
    IMAGING: SCIMAGING,
    SPECTROSCOPIC: SCSPECTROSCOPIC
}


def make_rectangular_img_snr_region(args, instrument):
    """ Makes a specialized imaging extraction region
        used in scan mode.

    Parameters
    ----------
    args: tuple or list
      with the region dimesnions and units
    instrument: an instance of Imager
      the instrument instance

    Returns
    -------
    instance of ImageRectangularArea

    """
    width = args[1][0]
    height = args[1][1]
    units = args[2]

    arcsec_per_pixel_height = instrument.detector.pixel_angheight
    arcsec_per_pixel_width = instrument.detector.pixel_angwidth

    rgn = ImageRectangularArea(SNR_REGION,
                               width,
                               height,
                               units,
                               arcsec_per_pixel_width,
                               arcsec_per_pixel_height)
    return rgn


def get_scan_configuration_data(ei):
    """ Loads instrument and detector data from config files.

        This function calls the base function and then loads
        scan mode parameters into the custom configuration dict.

    Parameters
    ----------
    ei: dict
      engine input

    Returns
    -------
    iconfig: dict
      dictionary of instrument data
    dconfig: dict
      dictionary of detector data
    custom_config: dict
      dictionary with custom data

    """
    # get baseline configuration data.
    iconfig, dconfig, custom_config = initialize.get_configuration_data(ei)

    # scan mode parameters
    custom_config['scan_mode'] = ei['science_mode'] in (SCIMAGING, SCSPECTROSCOPIC)
    if custom_config['scan_mode']:
        custom_config['scan_length'] = ei['sclength']

    return iconfig, dconfig, custom_config


def modify_scan_configuration(instrument, custom_config):
    """ The scan_mode flag must be propagated into the instrument
        instance. Science mode must be translated properly, since
        the instruments by default only recognize 'imaging' and
        'spectroscopic' as valid modes. As a consequence, the
        limits config file must be revisited again.

    Parameters
    ----------
    instrument: Imager or Spectrograph
      Instance being configured
    custom_config: dictionary
      custom configuration info

    """
    if custom_config['scan_mode']:

        instrument.scan_mode = True
        instrument.scan_length = custom_config['scan_length']
        instrument.science_mode = scan_science_modes[instrument.science_mode]

        # In scan mode, the limits config file must be revisited because
        # the science mode was just redefined by statement above. The
        # scan mode-specific limits may have been missed by the default
        # read that happened in the generic instrument._configure method.
        instrument.set_instrument_detector_limits()



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
import scipy.interpolate as sc
from photutils.geometry import circular_overlap_grid, elliptical_overlap_grid

from .custom_exceptions import EngineInputError


class Grid:

    """
    Generate an x, y grid in a rectangular region, sampled with xsamp and
    ysamp spacings in the x and y directions, respectively.  Origin is at the
    upper-left corner to match numpy and matplotlib convention. astropy.io.fits
    also assumes UL origin by default.

    What this means, practically, is that all convolution or other operations 
    that affect the datacube or projected array directly (i.e., without using 
    the Grid class) have to be indexed as (y,x), in order to be properly 
    oriented. All functions (e.g. placing stars, dropping extraction masks) 
    that use this Grid class or items derived from it can access it with 
    (x,y) coordinates.

    Parameters
    ----------
    xsamp, ysamp : float, float
        The sampling spacing in the x and y directions.
    nx, ny: int, int
        Number of samples in the x and y directions.
    """

    def __init__(self, xsamp, ysamp, nx, ny):
        self.xsamp = np.abs(xsamp)
        self.ysamp = np.abs(ysamp)
        startx = -(nx - 1) / 2.0 * self.xsamp
        stopx = (nx - 1) / 2.0 * self.xsamp
        starty = -(ny - 1) / 2.0 * self.ysamp
        stopy = (ny - 1) / 2.0 * self.ysamp
        xvals = np.linspace(startx, stopx, num=nx)
        yvals = np.linspace(starty, stopy, num=ny)

        ones = np.ones((ny, nx))
        x = ones * xvals
        y = np.flipud(ones * yvals.reshape(int(ny), 1))

        self.nx = nx
        self.ny = ny
        self.x = x
        self.y = y
        self.row = xvals
        # flip Y axis because we use Y increasing from bottom to top
        self.col = yvals[::-1]

        pixel_startx = -(nx - 1) / 2.0
        pixel_stopx = (nx - 1) / 2.0
        pixel_starty = -(ny - 1) / 2.0
        pixel_stopy = (ny - 1) / 2.0
        pixel_xvals = np.linspace(pixel_startx, pixel_stopx, num=nx)
        pixel_yvals = np.linspace(pixel_starty, pixel_stopy, num=ny)

        self.pixel_row = pixel_xvals
        # flip Y axis because we use Y increasing from bottom to top
        self.pixel_col = pixel_yvals[::-1]

    @property
    def shape(self):
        """
        Provides array-like .shape functionality
        """
        sh = (self.ny, self.nx)
        return sh

    def as_dict(self):
        """
        Create engine API format dict containing the properties of the Grid instance. The half-pixel
        offsets are due to the pixel coordinates being referenced to pixel centers. Thus the edges of the
        Grid are a half-pixel above and to the left of the Grid index origin.
        """

        transform = {}
        transform['x_refpix'] = 0
        transform['x_refval'] = self.x[0][0]
        transform['x_max'] = self.x.max() + 0.5 * self.xsamp
        transform['x_min'] = self.x.min() - 0.5 * self.xsamp
        transform['x_step'] = self.xsamp
        transform['x_size'] = self.x.shape[1]
        transform['y_refpix'] = 0
        transform['y_refval'] = self.y[0][0]
        transform['y_max'] = self.y.max() + 0.5 * self.ysamp
        transform['y_min'] = self.y.min() - 0.5 * self.ysamp
        transform['y_step'] = -self.ysamp
        transform['y_size'] = self.y.shape[0]
        return transform

    def wcs_info(self):
        """
        Define coordinate transform in WCS header format.

        Returns
        -------
        header: dict
            WCS keys defining coordinate transform for the 2 spatial axes
        """
        t = self.as_dict()
        header = {
            'ctype1': 'X offset',
            'crpix1': 1,
            'crval1': t['x_min'],
            'cdelt1': t['x_step'],
            'cunit1': 'arcsec',
            'cname1': 'X',
            'ctype2': 'Y offset',
            'crpix2': 1,
            'crval2': t['y_min'],
            'cdelt2': -t['y_step'],
            'cunit2': 'arcsec',
            'cname2': 'Y',
        }
        return header

    def __len__(self):
        """
        Provide array-like len() functionality
        """
        l = self.x.shape[0] * self.x.shape[1]
        return l

    def __getitem__(self, val):
        """
        Provide array-like indexing functionality.

        Parameters
        ----------
        val : slice object
            Valid python index or slice() specification.

        Returns
        -------
        pos : [y, x] list of floats or arrays
            Elements are floats in the case of specific indicies and arrays in
            the cases of slice()'s.
        """
        section = [self.y[val], self.x[val]]
        return section

    def bounds(self):
        return {'xmin': np.min(self.x),
                'xmax': np.max(self.x),
                'ymin': np.min(self.y),
                'ymax': np.max(self.y)}

    def dist(self, xcen=0.0, ycen=0.0):
        """
        Return a distance array where each element contains its distance from
        the center of the grid.
        """
        d = np.sqrt((self.x - xcen) ** 2 + (self.y - ycen) ** 2)
        return d

    def dist_xy(self, xcen=0.0, ycen=0.0):
        """
        Return a directional distance array where each element contains its (x,y) distance from
        the center of the grid.
        """
        dx = self.x - xcen
        dy = self.y - ycen
        return dx,dy

    def world_to_image(self, y, x):
        """
        Return fractional index coordinates yi,xi corresponding to given y,x position

        Parameters
        ----------
        y, x : float
            y, x position in world coordinates

        Returns
        -------
        pos : [yi, xi] list of floats
            yi, xi image coordinates that correspond to given position
        """
        t = self.as_dict()
        if x < t['x_min'] or x > t['x_max']:
            raise ValueError("X outside of Grid bounds.")
        if y < t['y_min'] or y > t['y_max']:
            raise ValueError("Y outside of Grid bounds.")
        x_image = x / self.xsamp + (self.shape[1] - 1) / 2.0
        # there's a minus sign here because y indices increase from up to down,
        # but y coordinates increase down to up.
        y_image = -y / self.ysamp + (self.shape[0] - 1) / 2.0
        pos = [y_image, x_image]
        return pos

    def world_to_index(self, y, x):
        """
        Return nearest index pair to given y,x position

        Parameters
        ----------
        y, x : float
            y,x position in world coordinates

        Returns
        -------
        pos : [y, x] list of ints
            y,x indices that most closely correspond to given position
        """
        y_index, x_index = self.world_to_image(y, x)
        y_index, x_index = self._index_bound(int(round(y_index)), self.ny), self._index_bound(int(round(x_index)), self.nx)
        pos = [y_index, x_index]
        return pos

    def _index_bound(self, index, nindex):
        """
        Apply bounds to indicies so the nearest index is either the first or last index for cases
        at edge or beyond Grid.

        Parameters
        ----------
        index: int
            Index to bounds-check
        nindex: int
            Length of axis being indexed

        Returns
        -------
        index: int
            Index with boundaries applied
        """
        if index < 0:
            index = 0
        if index > nindex - 1:
            index = nindex - 1
        return index

    def shift_rotate(self, yoff, xoff, rot):
        """
        Return shifted/rotated (y, x) given offsets (yoff, xoff) and rotation, rot (degrees)

        Parameters
        ----------
        yoff, xoff: float
            yoff, xoff offsets in world coordinates
        rot: float
            rotation angle in degrees

        Returns
        -------
        ysh_rot, xsh_rot: 2D numpy arrays
            rotated and shifted copies of Grid.x and Grid.y
        """
        pa_radians = np.pi * rot / 180.0
        xsh = self.x - xoff
        ysh = self.y - yoff
        xsh_rot = xsh * np.cos(pa_radians) + ysh * np.sin(pa_radians)
        ysh_rot = -xsh * np.sin(pa_radians) + ysh * np.cos(pa_radians)
        return ysh_rot, xsh_rot

    def get_aperture(self):
        """
        Return Grid parameters as an aperture specification
        """
        ap = {'width': self.nx * self.xsamp, 'height': self.ny * self.ysamp, 'offset': (0, 0)}
        return ap

    def rectangular_mask(self, width=None, height=None, xoff=0.0, yoff=0.0, transparency=1.0, pixels=False):
        """
        Define a rectangular mask within the Grid of specfied width and height and offset by xoff/yoff.
        Pixels partially covered by mask are properly weighted by the area subtended by the mask.

        Parameters
        ----------
        width: float
            Width of the mask. If None, then use the full width of the Grid.
        height: float
            Height of the mask. If None, then use the full height of the Grid.
        xoff: float
            X offset of the mask
        yoff: float
            Y offset of the mask
        transparency: float (default: 1.0)
            Transparency of the mask
        pixels: bool
            When False, the width and height are in units of arcseconds. When True, they are in units of 
            pixels. Offsets are always in arcseconds. 

        Returns
        -------
        mask: 2D np.ndarray
            2D mask image
        """
        if transparency < 0.0 or transparency > 1.0:
            msg = "Mask transparency, %f, must be in the range of 0.0 (fully opaque) to 1.0 (fully clear)." % transparency
            raise EngineInputError(value=msg)

        if pixels:
            minx = self.pixel_row.min()
            maxx = self.pixel_row.max()
            miny = self.pixel_col.min()
            maxy = self.pixel_col.max()
            xsamp = 1
            ysamp = 1
            # The offsets are still expected to be in arcseconds, so we need to convert
            # them to pixels for pixel usage.
            xoff /= self.xsamp
            yoff /= self.ysamp
        else:
            minx = self.row.min()
            maxx = self.row.max()
            miny = self.col.min()
            maxy = self.col.max()
            xsamp = self.xsamp
            ysamp = self.ysamp
        xcen = 0
        ycen = 0

        # if width and/or height are not specified, set them encompass the size of the Grid.
        # add two pixels to make sure the masking does not interpolate the edges.
        if width is None:
            width = maxx - minx + 2 * xsamp
        if height is None:
            height = maxy - miny + 2 * ysamp

        mask_x = np.array(
            [
                minx,
                xcen - (width + xsamp) / 2.0 + xoff,
                xcen - (width - xsamp) / 2.0 + xoff,
                xcen + (width - xsamp) / 2.0 + xoff,
                xcen + (width + xsamp) / 2.0 + xoff,
                maxx
            ]
        )
        mask_y = np.array(
            [
                miny,
                ycen - (height + ysamp) / 2.0 + yoff,
                ycen - (height - ysamp) / 2.0 + yoff,
                ycen + (height - ysamp) / 2.0 + yoff,
                ycen + (height + ysamp) / 2.0 + yoff,
                maxy
            ]
        )

        mask_pix = np.array([0.0, 0.0, 1.0, 1.0, 0.0, 0.0])
        if pixels:
            x_mask_int = np.interp(self.pixel_row, mask_x, mask_pix)
            y_mask_int = np.interp(self.pixel_col, mask_y, mask_pix)
        else:
            x_mask_int = np.interp(self.row, mask_x, mask_pix)
            y_mask_int = np.interp(self.col, mask_y, mask_pix)
        x_mask = x_mask_int.reshape((1, self.nx))
        y_mask = y_mask_int.reshape((self.ny, 1))
        mask = y_mask * x_mask * transparency
        return mask

    def circular_mask(self, radius, xoff=0.0, yoff=0.0, use_exact=1, subsampling=1, transparency=1.0, pixels=False):
        """
        Define a circular mask within the grid of a specified radius and offset by xoff/yoff.
        Use photutils.geometry to create masks that calculate the fraction of a pixel subtended
        by the circular mask.  use_exact=1 performs the exact geometric calculation while
        use_exact=0 will sub-sample the pixels by the specfied subsampling parameter
        to estimate the subtended area.

        Parameters
        ----------
        radius: float
            Radius of the circular mask
        xoff: float
            X position of the center of the mask
        yoff: float
            Y position of the center of the mask
        use_exact: int (default: 1)
            If 1, then use exact geometrical calculation to weight pixels partially covered by mask.
            This parameter is passed directly on to photutils.geometry.circular_overlap_grid()
        subsampling: int (default: 1)
            If use_exact=0, then subsample pixels by this factor to calculate area of each pixel subtended
            by the mask. The default value of 1 will force no partial pixels to be included in the mask.
            This parameter is passed directly on to photutils.geometry.circular_overlap_grid()
        transparency: float (default: 1.0)
            Transparency of the mask
        pixels: bool
            When False, the width and height are in units of arcseconds. When True, they are in units of 
            pixels. Offsets are always in arcseconds. 

        Returns
        -------
        mask: 2D np.ndarray
            2D mask image
        """
        if transparency < 0.0 or transparency > 1.0:
            msg = "Mask transparency, %f, must be in the range of 0.0 (fully opaque) to 1.0 (fully clear)." % transparency
            raise EngineInputError(value=msg)

        t = self.as_dict()
        # we need to use flipud because we use an origin in the UL corner of an image
        # while photutils uses the LL corner.
        if pixels:
            xmax = self.pixel_row.max() + 0.5
            xmin = self.pixel_row.min() - 0.5
            ymax = self.pixel_col.max() + 0.5
            ymin = self.pixel_col.min() - 0.5
            mask = np.flipud(
                # The circular overlap grid uses the extent in (units), and the size in
                # index positions, to determine a pixel scale to create the circular mask.
                circular_overlap_grid(
                    xmin - xoff / self.xsamp,
                    xmax - xoff / self.xsamp,
                    ymin - yoff / self.ysamp,
                    ymax - yoff / self.ysamp,
                    t['x_size'],
                    t['y_size'],
                    radius,
                    use_exact,
                    subsampling
                )
            )
        else:
            mask = np.flipud(
                circular_overlap_grid(
                    t['x_min'] - xoff,
                    t['x_max'] - xoff,
                    t['y_min'] - yoff,
                    t['y_max'] - yoff,
                    t['x_size'],
                    t['y_size'],
                    radius,
                    use_exact,
                    subsampling
                )
            )
        mask *= transparency
        return mask

    def elliptical_mask(self, major, minor, pa=0.0, xoff=0.0, yoff=0.0, use_exact=1, subsampling=1, transparency=1.0):
        """
        Define an elliptical mask within the grid with specified major/minor axes, position angle,
        and offset from center by xoff/yoff. Use photutils.geometry to create masks that calculate
        the fraction of a pixel subtended by the mask.  use_exact=1 performs the exact geometric
        calculation while use_exact=0 will sub-sample the pixels by the specfied subsampling parameter
        to estimate the subtended area.

        Parameters
        ----------
        major: float
            Semi-major axis of the elliptical mask
        minor: float
            Semi-minor axis of the mask
        pa: float
            Position angle of the ellipse in degrees measured clockwise (positive in +X direction)
        xoff: float
            X position of the center of the mask
        yoff: float
            Y position of the center of the mask
        use_exact: int (default: 1)
            If 1, then use exact geometrical calculation to weight pixels partially covered by mask.
            This parameter is passed directly on to photutils.geometry.circular_overlap_grid()
        subsampling: int (default: 1)
            If use_exact=0, then subsample pixels by this factor to calculate area of each pixel subtended
            by the mask. The default value of 1 will force no partial pixels to be included in the mask.
            This parameter is passed directly on to photutils.geometry.circular_overlap_grid()
        transparency: float (default: 1.0)
            Transparency of the mask

        Returns
        -------
        mask: 2D np.ndarray
            2D mask image
        """
        if transparency < 0.0 or transparency > 1.0:
            msg = "Mask transparency, %f, must be in the range of 0.0 (fully opaque) to 1.0 (fully clear)." % transparency
            raise EngineInputError(value=msg)

        theta = np.pi * pa / 180.0  # photutils uses angles in radians

        t = self.as_dict()
        # we need to use flipud because we use an origin in the UL corner of an image
        # while photutils uses the LL corner.
        mask = np.flipud(
            elliptical_overlap_grid(
                t['x_min'] - xoff,
                t['x_max'] - xoff,
                t['y_min'] - yoff,
                t['y_max'] - yoff,
                t['x_size'],
                t['y_size'],
                major,
                minor,
                theta,
                use_exact,
                subsampling
            )
        )
        mask *= transparency
        return mask

    def point_source(self, xoff=0.0, yoff=0.0):
        """
        Define a point source within the grid at position xoff, yoff.  Sub-pixel positioning
        of the point source is performed by weighting the 4 nearest pixels.
        Light will be split among 1, 2, or 4 pixels in a 2x2 box depending on how it's centered. We calculate:
        a.) the lower left pixel of the 2x2 box,
        b.) the proportion of flux that spills into that and the other three pixels (as long as they're not off the edge
            of the grid)
        c.) and then return an image, scaled appropriately.

        Parameters
        ----------
        xoff: float
            X position of point source
        yoff: float
            Y position of point source

        Returns
        -------
        im: 2D np.ndarray
            Image containing point source with flux scaled to 1.0.
        """
        im = np.zeros(self.shape)

        # normally it's an error to specify a position off the Grid, but in this case it just
        # means there's no flux being added onto the image. in cases such as dithering, putting
        # the point source off the Grid is intentional and shouldn't generate errors or warnings.
        try:
            yi, xi = self.world_to_image(yoff, xoff)
        except ValueError as e:
            return im

        rem_x, int_x = np.modf(xi)
        rem_y, int_y = np.modf(yi)

        # check for image edges and handle weights correctly if point source straddles an edge
        if rem_x < 0.0:
            x_shift = -1
            rem_x = 1.0 + rem_x
        else:
            x_shift = 0

        if rem_y < 0.0:
            y_shift = -1
            rem_y = 1.0 + rem_y
        else:
            y_shift = 0

        # here we handle the cases where a point source is at the edge of the Grid so that some of the
        # flux is in pixels on the Grid, while the rest is off the Grid.
        ly = int(int_y + y_shift)
        uy = int(ly + 1)
        lx = int(int_x + x_shift)
        ux = int(lx + 1)

        # if a pixel is None, then it's off the Grid and we don't try to set it.
        if ly < 0:
            ly = None
        if uy == self.ny:
            uy = None
        if lx < 0:
            lx = None
        if ux == self.nx:
            ux = None

        if ly is not None and lx is not None:
            im[ly, lx] += (1.0 - rem_x) * (1.0 - rem_y)
        if uy is not None and lx is not None:
            im[uy, lx] += (1.0 - rem_x) * rem_y
        if ly is not None and ux is not None:
            im[ly, ux] += rem_x * (1.0 - rem_y)
        if uy is not None and ux is not None:
            im[uy, ux] += rem_x * rem_y

        return im


class IrregularGrid(Grid):

    """
    Create Grid from pre-specified rows and cols. Uniform spacing is still assumed.
    """

    def __init__(self, yvals, xvals):
        ny = yvals.size
        nx = xvals.size
        ones = np.ones((ny, nx))
        y = ones * yvals.reshape(ny, 1)
        x = ones * xvals

        self.nx = nx
        self.ny = ny
        self.x = x
        self.y = y
        self.row = xvals
        self.col = yvals
        if len(xvals) > 1:
            self.xsamp = np.abs(xvals[1] - xvals[0])
        if len(yvals) > 1:
            self.ysamp = np.abs(yvals[1] - yvals[0])

        # For pixels, we just need to find the offset of 0 and count from there.
        xzero = np.interp(0, xvals, np.arange(nx))
        yzero = np.interp(0, yvals[::-1], np.arange(ny))    

        self.pixel_row = np.arange(nx) - xzero
        # flip Y axis because we use Y increasing from bottom to top
        self.pixel_col = np.arange(ny)[::-1] - yzero

    def world_to_image(self, y, x):
        """
        Return fractional index coordinates yi,xi corresponding to given y,x position
        This version can handle irregular grids (though with position found via linear
        interpolation)

        Parameters
        ----------
        y, x : float
            y, x position in world coordinates

        Returns
        -------
        pos : [yi, xi] list of floats
            yi, xi image coordinates that correspond to given position
        """
        t = self.as_dict()
        if x < t['x_min'] or x > t['x_max']:
            raise ValueError("X outside of Grid bounds.")
        if y < t['y_min'] or y > t['y_max']:
            raise ValueError("Y outside of Grid bounds.")
        x_interp = sc.interp1d(self.row, list(range(self.nx)))
        x_image = x_interp(x)
        #x_image = np.interp(x, self.row, list(range(self.nx)))
        # y coordinates are flipped, requiring the lists to be flipped
        y_interp = sc.interp1d(self.col, list(range(self.ny)))
        y_image = y_interp(y)
        #y_image = np.interp(y, self.col[::-1], list(range(self.ny))[::-1])
        pos = [float(y_image), float(x_image)]

        return pos



#! /usr/bin/env python
"""
This is a collection of pyetc utils and stubs to get the
pyetc code working (minimally) under pandeia as needed.
"""

import os, pprint, re, sys
import numpy as np
from numpy import nan, inf, NaN, Inf
from collections import OrderedDict
import ast

class mock_log:
    """ Mock up a "log" pkg that simply matches what the old pyetc code needs.
        The intent is to NOT bring in any more pyetc code than necessary. """

    def __init__(self, be_verbose=False):
        self.VERBOSE_FLAG = be_verbose

    def info(self, str1, str2='', str3=''):
        print(str1, str2, str3)

    def error(self, str1, str2='', str3='', str4=''):
        print(str1, str2, str3, str4, file=sys.stderr)

    def write(self, str1='', eol='\n'):
        print(str1, end=eol, flush=True)

    verbose = info # oddly this is the way it works in pyetc; caller checks VERBOSE_FLAG
    warning = error # these seem to be the same basically

log = mock_log(False)

# rather than take PYSYN_CDBS from a pyetc config system, pull from env, but don't make a critical error on import
PYSYN_CDBS = os.environ.get('PYSYN_CDBS','PYSYN_CDBS_is_not_set' )

# pyetc's dynamic_import
PACKAGE_RE = re.compile("[A-Za-z_0-9.]+")
def dynamic_import(package):
    """imports a module specified by string `package` which is
    not known until runtime.

    Returns a module/package.

    The purpose of this function is to concentrate a number of uses
    of the "exec" statement in once place where it can be policed
    for security reasons.
    """
    if not PACKAGE_RE.match(package):
        raise ImportError("Invalid dynamic import " + repr(package))

    # Python 3 exec doesn't support keyword arguments
    exec("import " + package + " as pkg_name", None, locals())

    m = locals()['pkg_name']
    return m


def read_dict(fname):
    ''' read a python dictionary from a file that was written with write_dict. '''
    f=open(fname,'r', encoding="utf-8")
    datastr = f.read()
    f.close()
    # convert DOS file to Unix - otherwise the eval will fail
    datastr = datastr.replace('\r','')
    try :
        datadict = safe_eval(datastr)
    except Exception as e:
        print('EXCEPTION:',e)
        print('cannot eval data in file ',fname)
        raise
    return datadict


def write_dict(datadict, fname, header=None):
    '''write a python dictionary to a file in such a way that it
    can be read with read_dict.
    At present, this means using pprint, but encapsulating it
    here will make it easier if we decide to change the format as the
    project evolves.
    A string, or a list of strings, passed in the optional header keyword
    will be prepended with a # sign before being written to the file,
    which read_dict will ignore. This allows the files to be
    documented.
    '''

    fh=open(fname,'w')
    #First write commented-out header
    if header is not None:
        #Support either a list or a string
        if isinstance(header,list):
            for line in header:
                fh.write("# %s\n"%line)
        else:
            fh.write("# %s\n"%header)
    #Now write the data
    pprint.pprint(datadict,fh)
    fh.close()


# stolen/modified from spidring/metadict.py (see code there as to why this is this way)
def fast_write_dict(adict, fname, comment):
    f = open(fname, 'w')
    f.write("# "+ comment + "\n")
    f.write("{\n")
    for key,val in sorted(adict.items()):
        f.write("\t" + repr(key) + " : " + repr(val) + ",\n")
    f.write("}\n")
    f.close()


# pyetc took from Python-2.7 and enhanced to support np.nan and np.inf
def safe_eval(node_or_string):
    """
    Safely evaluate an expression node or a string containing a Python
    expression.  The string or node provided may only consist of the following
    Python literal structures: strings, numbers, tuples, lists, dicts, booleans,
    and None.
    """
    _safe_names = {'None': None, 'True': True, 'False': False,
                   'nan':nan, 'NaN':np.NaN,
                   'inf':inf, 'Inf':np.Inf}
    if isinstance(node_or_string, str):
        node_or_string = ast.parse(node_or_string, mode='eval')
    if isinstance(node_or_string, ast.Expression):
        node_or_string = node_or_string.body
    def _convert(node):
        # Old code used to return instances of ast.Dict.
        # New code sometimes retuns dicts disguised as lists
        # of tuples (actuaally, ast.List and ast.Tuple), as
        # the first element of the 'args' attribute of an
        # ast.Call instace.
        if isinstance(node, ast.Call):
            result = OrderedDict()
            if len(node.args) > 0:
                for element in node.args[0].elts:
                    name  = element.elts[0].s
                    v = element.elts[1]
                    if isinstance(v, ast.Str):
                        value = v.s
                    elif isinstance(v, ast.Num):
                        value = v.n
                    else:
                        value = None
                    result[name] = value
            return result
        elif isinstance(node, ast.Str):
            return node.s
        elif isinstance(node, ast.Num):
            return node.n
        elif isinstance(node, ast.Tuple):
            return tuple(map(_convert, node.elts))
        elif isinstance(node, ast.List):
            return list(map(_convert, node.elts))
        elif isinstance(node, ast.Dict):
            return OrderedDict((_convert(k), _convert(v)) for k, v
                         in zip(node.keys, node.values))
        elif isinstance(node, ast.Name):
            if node.id in _safe_names:
                return _safe_names[node.id]
        elif isinstance(node, ast.BinOp) and \
             isinstance(node.op, (ast.Add, ast.Sub)) and \
             isinstance(node.right, ast.Num) and \
             isinstance(node.right.n, complex) and \
             isinstance(node.left, ast.Num) and \
             isinstance(node.left.n, int) and \
             isinstance(node.left.n, float):
            left = node.left.n
            right = node.right.n
            if isinstance(node.op, ast.Add):
                return left + right
            else:
                return left - right

        # when the expression contains a negative number,
        # under python 3 the negative number will be cast
        # as an instance of UnaryOp with first operand of
        # type USub.
        elif isinstance(node, ast.NameConstant):
            if sys.version_info[0] >= 3:
                return node.value
        elif isinstance(node, ast.UnaryOp) and \
             isinstance(node.op, ast.USub) and \
             isinstance(node.operand, ast.Num):
            return -(node.operand.n)

        raise ValueError('malformed string')

    return _convert(node_or_string)


def stable_dict(dictionary=None):
    ''' Returns a dictionary of the appropriate type.

        Under python 2, dictionary iterators are stable
        in between successivea calls within the same process,
        or from process to procees. Under python 3, dictionary
        iterators return elements in random order, different
        from call to call. This function returns an OrderedDict
        instance that works under both 2 and 3, but leaves open
        the possibility to use other kinds of dictionary if the
        need ever arises.

        This function should be used to create dictionary instances
        throughout the pyetc code, instead of direct calls to dict().
    '''
    if dictionary:
        return OrderedDict([item for item in dictionary.items()])
    else:
        return OrderedDict()
